"""
------------------------------------------------------------------------------------------
pygame实现游戏“坦克大战”
2021.6.15
------------------------------------------------------------------------------------------
"""

import random
import sys
import time
# from _curses import COLOR_GREEN
from urllib.request import urlretrieve
import os
import pygame
from pygame.locals import *
from pygame.sprite import Sprite
import numpy as np

'''全局变量'''
SCREEN_WIDTH, SCREEN_HEIGHT = 1200, 700  # 画面大小
CREAT_ENEMY_TANK = pygame.USEREVENT  # 监听事件

TANK_SPEED = [5, 10, 15, 20]  # 坦克的速度
BULLET_RADIATION = [4, 9, 14, 19]  # 子弹射程
ENEMY_TANK_COUNT = [5, 6, 7, 8]

COLOR_BLACK = pygame.Color(0, 0, 0)
COLOR_WHITE = pygame.Color(255, 255, 255)
COLOR_RED = pygame.Color(255, 0, 0)

'''定义一个精灵类'''


class BaseItem(Sprite):
    def __int__(self, color, width, height):
        pygame.sprite.Sprite.__init__(self)


'''坦克主类'''


class Tank(BaseItem):
    # 定义所有坦克的性质：长宽大小
    width = 60
    height = 60

    def __init__(self, left, top):
        self.direction = 'U'

        # 存放图片的字典
        self.images = {
            'U': pygame.image.load('img/m_U.png'),
            'D': pygame.image.load('img/m_D.png'),
            'L': pygame.image.load('img/m_L.png'),
            'R': pygame.image.load('img/m_R.png'),
        }
        # 由方向字典决定坦克的图片
        self.image = self.images[self.direction]
        self.speed = TANK_SPEED[0]  # 坦克的速度
        self.rect = self.image.get_rect()

        # 坦克的位置
        self.rect.left = left
        self.rect.top = top
        self.stop = True  # 坦克是否停下
        self.live = True  # 坦克是否被消灭

        # 记录原来的位置
        self.prevLeft = self.rect.left
        self.prevTop = self.rect.top

    # 射击
    def shot(self):
        return Bullet(self)

    # 移动
    def tankMove(self):
        # 保持原来状态
        self.prevTop = self.rect.top
        self.prevLeft = self.rect.left
        # 移动
        if self.direction == 'U':  # 上移
            if self.rect.top >= 0:
                self.rect.top -= self.speed
        elif self.direction == 'D':  # 下移
            if self.rect.top + self.rect.height <= SCREEN_HEIGHT:
                self.rect.top += self.speed
                # else:
                #   self.rect.top = SCREEN_HEIGHT  #移到边缘
        elif self.direction == 'L':  # 左移
            if self.rect.left >= 0:
                self.rect.left -= self.speed
        elif self.direction == 'R':  # 右移
            if self.rect.left + self.rect.height <= SCREEN_WIDTH:
                self.rect.left += self.speed
                # else:
                #   self.rect.left = SCREEN_WIDTH

    # 位置保持不变
    def tankStay(self):
        self.rect.top = self.prevTop
        self.rect.left = self.prevLeft

    # 加载坦克
    def displayTank(self):
        self.image = self.images[self.direction]
        MainGame.window.blit(self.image, self.rect)

    # 撞墙
    def hitWall(self):
        for wall in MainGame.wallList:
            if pygame.sprite.collide_rect(wall, self):
                self.tankStay()


'''己方坦克'''


class MyTank(Tank):
    def __init__(self, left, top):
        super().__init__(left, top)

    def myTank_hit_enemyTank(self):
        for enemyTank in MainGame.EnemyTankList:
            if pygame.sprite.collide_rect(enemyTank, self):
                self.tankStay()


'''敌方坦克'''


class EnemyTank(Tank):
    def __init__(self, left, top, speed):
        super(EnemyTank, self).__init__(left, top)
        # 敌方坦克方向和图片字典
        self.images = {
            'U': pygame.image.load('img/e_U.png'),
            'D': pygame.image.load('img/e_D.png'),
            'L': pygame.image.load('img/e_L.png'),
            'R': pygame.image.load('img/e_R.png'),
        }

        # 随机生成敌方坦克方向
        self.direction = self.RandomDirection()
        # 由方向字典决定坦克的图片
        self.image = self.images[self.direction]
        self.speed = speed  # 坦克的速度
        self.rect = self.image.get_rect()

        # 坦克的位置
        self.rect.left = left
        self.rect.top = top
        # self.IsTurned = False
        self.IsEnemy = True
        self.step = 30

    # 生成随机方向
    def RandomDirection(self):
        num = random.randint(1, 4)
        if num == 1:
            return 'U'
        elif num == 2:
            return 'D'
        elif num == 3:
            return 'L'
        else:
            return 'R'

    # 进行随机移动
    def RandomMove(self):
        if self.step <= 0:
            self.direction = self.RandomDirection()
            self.step = 30
        else:
            self.tankMove()
            self.step -= 1

    # 进行随机射击
    def RandomShot(self):
        num = random.randint(1, 300)
        if num < 10:
            return Bullet(self)

    # 敌方坦克撞击己方坦克
    def enemyTank_hit_MyTank(self):
        for enemy in MainGame.EnemyTankList:
            if MainGame.my_tank and MainGame.my_tank.live:
                if pygame.sprite.collide_rect(MainGame.my_tank, enemy):
                    # 撞到己方坦克则停下
                    self.tankStay()


'''子弹类'''


class Bullet(BaseItem):
    def __init__(self, tank):
        # 存放图片的字典
        self.images = {
            'U': pygame.image.load('img/my_bullet_U.png'),
            'D': pygame.image.load('img/my_bullet_D.png'),
            'L': pygame.image.load('img/my_bullet_L.png'),
            'R': pygame.image.load('img/my_bullet_R.png'),
        }
        self.speed = 10
        self.live = True
        # 根据坦克位置加载子弹位置
        self.direction = tank.direction
        self.image = self.images[self.direction]
        self.rect = self.image.get_rect()

        if self.direction == 'U':
            self.rect.left = tank.rect.left + tank.rect.width / 2 - self.rect.width / 2
            self.rect.top = tank.rect.top - self.rect.height
        elif self.direction == 'D':
            self.rect.left = tank.rect.left + tank.rect.width / 2 - self.rect.width / 2
            self.rect.top = tank.rect.top + tank.rect.height
        elif self.direction == 'L':
            self.rect.left = tank.rect.left - self.rect.width / 2 - self.rect.width / 2
            self.rect.top = tank.rect.top + tank.rect.height / 2 - self.rect.width / 2
        elif self.direction == 'R':
            self.rect.left = tank.rect.left + tank.rect.width
            self.rect.top = tank.rect.top + tank.rect.height / 2 - self.rect.width / 2

    def displayBullet(self):
        MainGame.window.blit(self.image, self.rect)

    def bulletMove(self):
        if self.direction == 'U':
            if self.rect.top > 0:
                self.rect.top -= self.speed
            else:
                self.live = False
        elif self.direction == 'R':
            if self.rect.left + self.rect.width < SCREEN_WIDTH:
                self.rect.left += self.speed
            else:
                self.live = False
        elif self.direction == 'D':
            if self.rect.top + self.rect.height < SCREEN_HEIGHT:
                self.rect.top += self.speed
            else:
                self.live = False
        elif self.direction == 'L':
            if self.rect.left > 0:
                self.rect.left -= self.speed
            else:
                self.live = False

    # 己方坦克击中敌方坦克
    def myHitEnemy(self):
        for enemyTank in MainGame.EnemyTankList:
            if pygame.sprite.collide_rect(enemyTank, self):
                enemyTank.live = False
                self.live = False

                explode = Explode(enemyTank)
                MainGame.explodeList.append(explode)

    def enemyHitMy(self):
        if MainGame.my_tank and MainGame.my_tank.live:
            if pygame.sprite.collide_rect(MainGame.my_tank, self):
                MainGame.my_tank.live = False
                self.live = False

                # 创建爆炸对象
                explode = Explode(MainGame.my_tank)
                MainGame.explodeList.append(explode)


'''墙类'''


class Wall:
    def __init__(self, left, top):
        self.image = pygame.image.load('img/brick_wall.png')
        self.rect = self.image.get_rect()
        self.rect.left = left
        self.rect.top = top

        self.live = True  # 墙是否存在
        self.thick = 1  # 墙体厚度

    # 加载墙体
    def displayWall(self):
        if self.live:
            MainGame.window.blit(self.image, self.rect)

    # 墙体被射击
    def wallGotShot(self):
        for Bullet in MainGame.myBulletList:
            if pygame.sprite.collide_rect(self, Bullet):
                self.thick -= 1
                Bullet.live = False
                if self.thick == 0:
                    self.live = False
                    # 创建爆炸对象
                    explode = Explode(self)
                    MainGame.explodeList.append(explode)

        for Bullet in MainGame.enemyBulletList:
            if pygame.sprite.collide_rect(self, Bullet):
                self.thick -= 1
                Bullet.live = False
                if self.thick == 0:
                    self.live = False
                    # 创建爆炸对象
                    explode = Explode(self)
                    MainGame.explodeList.append(explode)


class Bricks(Wall):
    def __init__(self, left, top):
        self.image = pygame.image.load('img/brick_wall.png')
        self.rect = self.image.get_rect()
        self.rect.left = left
        self.rect.top = top

        self.live = True  # 墙是否存在
        self.thick = 1  # 墙体厚度


class Steel(Wall):
    def __init__(self, left, top):
        self.image = pygame.image.load('img/steels.gif')
        self.rect = self.image.get_rect()
        self.rect.left = left
        self.rect.top = top

        self.live = True  # 墙是否存在
        self.thick = np.inf  # 墙体厚度


class Camp(Wall):
    def __init__(self, left, top):
        self.image = pygame.image.load('img/camp.png')
        self.rect = self.image.get_rect()
        self.rect.left = left
        self.rect.top = top

        self.live = True  # 墙是否存在
        self.thick = 1  # 墙体厚度

    def campIsLive(self):
        if self.live == True:
            MainGame.active = True
        else:
            MainGame.active = False


'''爆炸类'''


class Explode:
    def __init__(self, tank):
        self.rect = tank.rect
        self.images = [
            pygame.image.load('img/blast0.gif'),
            pygame.image.load('img/blast1.gif'),
            pygame.image.load('img/blast2.gif'),
            pygame.image.load('img/blast3.gif'),
            pygame.image.load('img/blast4.gif')
        ]
        self.distance = 0
        self.image = self.images[self.distance]
        self.live = True

    # 加载爆炸类
    def displayExplode(self):
        if self.distance < len(self.images):
            self.image = self.images[self.distance]
            self.distance += 1
            MainGame.window.blit(self.image, self.rect)
        else:
            self.live = False
            self.distance = 0


'''音乐播放类'''


class Music():
    def __init__(self, fileName):
        self.fileName = fileName
        # 先初始化混合器
        pygame.mixer.init()
        pygame.mixer.music.load(self.fileName)

    # 开始播放音乐
    def play(self):
        pygame.mixer.music.play()
        pass


'''按钮类'''


class Button(object):
    def __init__(self, screen, msg):
        """初始化按钮的属性"""
        self.screen = screen
        self.screen_rect = screen.get_rect()

        # 设置按钮的尺寸和其他属性
        self.width, self.height = 200, 50
        self.button_color = (0, 255, 0)
        self.text_color = (255, 255, 255)
        # 初始化字体模块
        pygame.font.init()
        self.font = pygame.font.SysFont('kaiti', 48)

        # 创建按钮的rect对象，居中
        self.rect = pygame.Rect(0, 0, self.width, self.height)
        self.rect.center = self.screen_rect.center

        # 按钮的标签只需创建一次
        self.prep_msg(msg)

    def prep_msg(self, msg):
        # 将msg渲染为图像
        self.msg_image = self.font.render(msg, True, self.text_color, self.button_color)
        # 使文字在按钮上居中
        self.msg_image_rect = self.msg_image.get_rect()
        self.msg_image_rect.center = self.rect.center

    def draw_button(self):
        # 初始化字体模块
        pygame.font.init()
        self.font = pygame.font.SysFont('kaiti', 48)
        # 绘制一个用颜色填充的按钮，再绘制文本
        self.screen.fill(self.button_color, self.rect)
        self.screen.blit(self.msg_image, self.msg_image_rect)


'''主进程'''


class MainGame:
    window = None
    my_tank = None
    camp = None
    play_button = None

    EnemyTankList = []
    EnemyTankCount = ENEMY_TANK_COUNT[0]
    EnemyDisplayCount = 0
    destory_enemy_tank = 0

    # 子弹列表
    myBulletList = []
    enemyBulletList = []

    explodeList = []

    wallList = []

    # 游戏开始及结束标志
    active = False
    end = False

    # 创建窗口
    def creatWindow(self):
        if not MainGame.window:
            MainGame.window = pygame.display.set_mode((SCREEN_WIDTH, SCREEN_HEIGHT))
        return MainGame.window

    # 创建开始按钮
    def createButton(self):
        MainGame.play_button = Button(MainGame.window, "PLAY")

    def createCamp(self):
        MainGame.camp = Camp(SCREEN_WIDTH / 2 - 30, SCREEN_HEIGHT - 60)
        MainGame.wallList.append(self.camp)

    # 创建己方坦克
    def createMyTank(self):
        MainGame.my_tank = MyTank(400, 300)

    def createEnemyTank(self):
        top = 0
        left = random.randint(0, 750)
        speed = 5  # random.randint(1, 5)
        enemy = EnemyTank(left, top, speed)
        MainGame.EnemyTankList.append(enemy)
        MainGame.EnemyDisplayCount += 1

    # 创建墙壁
    def createWall(self):
        for i in range(60, int(SCREEN_WIDTH / 2) - 60, 120):
            top = SCREEN_WIDTH // 3
            left = i
            wall = Bricks(left, top)
            MainGame.wallList.append(wall)

        for i in range(600, int(SCREEN_WIDTH / 2) - 60, 120):
            top = SCREEN_WIDTH // 5
            left = i
            wall = Bricks(left, top)
            MainGame.wallList.append(wall)

        for i in range(60, 400, 60):
            top = top = SCREEN_WIDTH // 9
            left = i
            wall = Bricks(left, top)
            MainGame.wallList.append(wall)

        for i in range(60, 400, 60):
            left = 900
            top = i
            wall = Bricks(left, top)
            MainGame.wallList.append(wall)

        for i in range(60, 400, 60):
            left = 700
            top = i
            wall = Bricks(left, top)
            MainGame.wallList.append(wall)

        # 加载大本营周围的保护墙
        brick1 = Bricks(MainGame.camp.rect.left - 60, MainGame.camp.rect.top)
        MainGame.wallList.append(brick1)
        brick2 = Bricks(MainGame.camp.rect.right, MainGame.camp.rect.top)
        MainGame.wallList.append(brick2)
        brick3 = Bricks(MainGame.camp.rect.right, MainGame.camp.rect.top - 60)
        MainGame.wallList.append(brick3)
        brick4 = Bricks(MainGame.camp.rect.left, MainGame.camp.rect.top - 60)
        MainGame.wallList.append(brick4)
        brick5 = Bricks(MainGame.camp.rect.left - 60, MainGame.camp.rect.top - 60)
        MainGame.wallList.append(brick5)

        # 金属墙壁
        for i in range(400, 700, 100):
            left = i
            top = 60
            wall = Steel(left, top)
            MainGame.wallList.append(wall)

        for i in range(500, 700, 60):
            left = i
            top = 500
            wall = Steel(left, top)
            MainGame.wallList.append(wall)

    # 加载墙壁
    def biltwall(self):
        for wall in MainGame.wallList:
            if wall.live:
                wall.displayWall()
                wall.wallGotShot()
            else:
                MainGame.wallList.remove(wall)

    # 加载敌方坦克
    def biltEnemyTank(self):
        for enemyTank in MainGame.EnemyTankList:
            if enemyTank.live:
                enemyTank.displayTank()
                enemyTank.RandomMove()
                enemyBullet = enemyTank.RandomShot()
                enemyTank.hitWall()
                enemyTank.enemyTank_hit_MyTank()

                if enemyBullet:
                    MainGame.enemyBulletList.append(enemyBullet)

            else:
                MainGame.EnemyTankList.remove(enemyTank)
                self.destory_enemy_tank += 1

    # 加载我方子弹
    def biltMyBullet(self):
        for bullet in MainGame.myBulletList:
            if bullet.live:
                bullet.displayBullet()
                bullet.bulletMove()
                bullet.myHitEnemy()
            else:
                MainGame.myBulletList.remove(bullet)

    # 加载敌方子弹
    def biltEnemyBullet(self):
        for bullet in MainGame.enemyBulletList:
            if bullet.live:
                bullet.displayBullet()
                bullet.bulletMove()
                bullet.enemyHitMy()

            else:
                MainGame.enemyBulletList.remove(bullet)

    # 加载爆炸效果
    def blitExplode(self):
        for explode in MainGame.explodeList:
            if explode.live:
                explode.displayExplode()
                music = Music('img/hit.wav')
                music.play()
            else:
                MainGame.explodeList.remove(explode)

    # 获取游戏中的所有事件
    def getEvent(self):
        # 获取游戏中的事件列表
        even_list = pygame.event.get()
        # 连续按键
        # pygame.key.set_repeat(5, 10)
        for e in even_list:
            # 点击窗口的叉号实现游戏结束
            if e.type == pygame.QUIT:
                sys.exit()

            # 通过上下左右键控制坦克的移动
            # 按键按下
            if e.type == pygame.KEYDOWN:
                if MainGame.my_tank and MainGame.my_tank.live:
                    if e.key == pygame.K_DOWN or e.key == pygame.K_s:
                        pygame.key.set_repeat(5, 10)
                        MainGame.my_tank.direction = 'D'
                        MainGame.my_tank.stop = False
                        print("按下向下的键，向下移动")
                    elif e.key == pygame.K_UP or e.key == pygame.K_w:
                        pygame.key.set_repeat(5, 10)
                        MainGame.my_tank.direction = 'U'
                        MainGame.my_tank.stop = False
                        print("按下向上的键，向上移动")
                    elif e.key == pygame.K_LEFT or e.key == pygame.K_a:
                        pygame.key.set_repeat(5, 10)
                        MainGame.my_tank.direction = 'L'
                        MainGame.my_tank.stop = False
                        print("按下向左的键，向左移动")
                    elif e.key == pygame.K_RIGHT or e.key == pygame.K_d:
                        pygame.key.set_repeat(5, 10)
                        MainGame.my_tank.direction = 'R'
                        MainGame.my_tank.stop = False
                        print("按下向右的键，向右移动")

                    elif e.key == pygame.K_SPACE:
                        pygame.key.set_repeat()
                        print('发射子弹')
                        # 创建我方子弹
                        if len(MainGame.myBulletList) < 10:
                            mybullet = Bullet(MainGame.my_tank)
                            MainGame.myBulletList.append(mybullet)

                            # 射击音效
                            Shot_music = Music('img/fire.wav')
                            Shot_music.play()

            # 鼠标相应事件
            elif e.type == pygame.MOUSEBUTTONDOWN:
                mouse_x, mouse_y = pygame.mouse.get_pos()
                # print(mouse_x)
                if MainGame.active == False:
                    if mouse_x >= MainGame.play_button.rect.left and mouse_x <= MainGame.play_button.rect.right:  # mouse_y <= MainGame.play_button.rect.top and mouse_y >= MainGame.play_button.rect.bottom:
                        MainGame.active = True
                        MainGame.myBulletList.clear()
                        MainGame.enemyBulletList.clear()
                        MainGame.EnemyTankList.clear()
                        MainGame.destory_enemy_tank = 0
                        MainGame.start_game.active = True
                        # print(MainGame.active)



            # 按键松开
            elif e.type == pygame.KEYUP:
                if e.key == pygame.K_UP or e.key == pygame.K_DOWN or e.key == pygame.K_LEFT or e.key == pygame.K_RIGHT \
                        or e.key == pygame.K_w or e.key == pygame.K_s or e.key == pygame.K_a or e.key == pygame.K_d:
                    if MainGame.my_tank and MainGame.my_tank.live:
                        MainGame.my_tank.stop = True

            # 监听事件：每隔1s添加敌方坦克
            elif e.type == CREAT_ENEMY_TANK:
                # print(0)
                # print(MainGame.EnemyDisplayCount, MainGame.EnemyTankCount)
                if MainGame.EnemyDisplayCount < MainGame.EnemyTankCount:
                    # print(MainGame.EnemyDisplayCount, MainGame.EnemyTankCount)
                    MainGame.createEnemyTank(MainGame)
                    # print(1)

    # 左上角文字绘制的功能
    def getTextSurface(self, text):
        # 初始化字体模块
        pygame.font.init()
        # 选一个合适的字体
        font = pygame.font.SysFont('kaiti', 18)
        # 使用对应的字符完成相关内容的绘制
        textSurface = font.render(text, True, COLOR_WHITE)
        return textSurface

    # 绘制开始按钮
    def drawButton(self):
        MainGame.play_button.draw_button()

    # 游戏结束提示语
    def endNote(self, text):
        # 初始化字体模块
        pygame.font.init()
        # 选中一个合适的字体
        font = pygame.font.SysFont('kaiti', 50)
        # 使用对应的字符完成相关内容的绘制
        textSurface = font.render(text, True, COLOR_RED)
        return textSurface

    # 游戏开始方法
    def start_game(self):
        # 初始化展示模块
        pygame.display.init()
        # 调用创建窗口的方法
        self.creatWindow()
        # 设置游戏窗口标题
        pygame.display.set_caption('坦克大战 by 全洧可')
        # 加载背景图片
        background = pygame.image.load('img/bkg3.jpg')

        # 程序持续进行
        while True:
            # 显示背景
            # 加载背景图片
            background = pygame.image.load('img/bkg3.jpg')
            self.window.blit(background, (0, 0))

            # 初始化开始按钮
            self.createButton()

            if not MainGame.active:
                MainGame.drawButton(MainGame.play_button)

            # 获取事件
            self.getEvent()

            # 初始化大本营
            self.createCamp()
            # 初始化我方坦克
            self.createMyTank()
            # 初始化敌方坦克
            self.createEnemyTank()
            # 初始化墙壁
            self.createWall()

            MainGame.EnemyDisplayCount = 0

            while MainGame.active:
                self.clock = pygame.time.Clock()
                # print(len(MainGame.EnemyTankList))
                # 显示背景
                self.window.blit(background, (0, 0))
                # 获取事件
                self.getEvent()
                # 将绘制文字得到的小画布，粘贴到窗口中
                MainGame.window.blit(self.getTextSurface("剩余敌方坦克%d辆" % len(MainGame.EnemyTankList)), (5, 5))
                MainGame.window.blit(self.getTextSurface("消灭敌方坦克%d辆" % self.destory_enemy_tank), (5, 20))
                # 背景音乐
                music = Music('img/start.wav')
                music.play()

                # 每隔1s加载一辆敌方坦克
                pygame.time.set_timer(CREAT_ENEMY_TANK, 1000)

                # 调用我方坦克进行显示
                if MainGame.my_tank and MainGame.my_tank.live:
                    MainGame.my_tank.displayTank()
                    if not MainGame.my_tank.stop:
                        MainGame.my_tank.tankMove()
                        MainGame.my_tank.hitWall()
                        MainGame.my_tank.myTank_hit_enemyTank()
                else:
                    del MainGame.my_tank
                    MainGame.my_tank = None
                # 加载我方子弹
                self.biltMyBullet()
                # 显示敌方坦克
                self.biltEnemyTank()
                # 显示敌方子弹
                self.biltEnemyBullet()
                # 显示墙壁
                self.biltwall()
                # 显示爆炸效果
                self.blitExplode()
                # 判断大本营摧毁
                self.camp.campIsLive()

                # self.addEnemyTank()

                # 窗口持续刷新
                pygame.display.update()
                time.sleep(0.02)

                if not self.my_tank.live:
                    self.active = False
                    self.end = True
                    MainGame.active = False
                    # continue

                if len(MainGame.EnemyTankList) == 0:
                    self.active = False
                    self.end = True
                    MainGame.active = False
                    # continue

                if not self.camp.live:
                    self.active = False
                    self.end = True
                    MainGame.active = False

            # 游戏结束，跳出提示信息
            if self.active == False and self.end == True:
                MainGame.window.blit(self.endNote("GAME IS OVER!"), (400, 100))
            # 窗口持续刷新
            pygame.display.update()
            time.sleep(0.02)
            # sys.exit(0)


pygame.init()
MainGame().start_game()
